/*
    Spakes - a client/server Snakes-clone for up to 4 players
    Copyright (C) 2001, 2002, 2004  Markus Triska

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact me at triska@gmx.at
*/

#include <stdio.h>
#include <sstream>
#include <algorithm>
#include <stdlib.h>

#include "spakes.h"
#ifdef WINDOWS
#include <time.h>
#include <winsock.h>
#include "window.h"
#else
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/time.h>
#include <unistd.h>
#include "gameboard.h"
#endif

#include "server.h"
#include "player.h"
#include "protocol.h"

using namespace std;

#include "mylog.h"
extern MyLogFile logfile;

int Server::read_byte(unsigned char &c)
{

	int numread;

	if (readpos >= writepos) {
		readpos = writepos = 0;
		fd_set allset;
		timeval t;
		t.tv_sec = 0;
		t.tv_usec = 0;
		FD_ZERO(&allset);
		FD_SET(servfd, &allset);
		if (select(servfd + 1, &allset, NULL, NULL, &t) < 0) {
			return -1;
		}
		if (!FD_ISSET(servfd, &allset)) {
			return -1;
		}

		if ( (numread = recv(servfd, data, 199, 0)) < 0) {
			return -1;
		} else if (numread == 0) {
			return 0;
		} else {
			writepos += numread;
		}
	}

	c = data[readpos++];
	return 1;
}


int Server::read_line(string &s)
{
	int retval;
	unsigned char oneletter;
	do {
		retval = read_byte(oneletter);
		if (retval <= 0) return retval;

		s += oneletter;
	} while (oneletter != DELIMINATER);

	s[s.size() - 1] = 0;

	return 1;
}



bool Server::receive_command(string &retstring)
{
	int numread = read_line(retstring);
	if (numread <= 0) return false;

	return true;
}

extern vector<Point> apples;
extern Player *players[4];
extern int numplayers;
extern bool multiplayer;
extern GameField field;

bool FreePos(Point &p)
{
	if (!field.free(p)) return false;
	return (find(apples.begin(), apples.end(), p) == apples.end());
}

#ifdef WINDOWS
extern Window *mainwindow;
#else
extern Gameboard *mainwindow;
#endif

bool Server::send_tick(Player &p) const
{
	if (p.Living()) {
		bool needsend = false;
		ostringstream command;
		int px = p.dots.back().x;
		int py = p.dots.back().y;
		bool loselife = false;
		if (p.outside()) {
			loselife = true;
		} else if (p.collision()) {
			if (p.num_ticks() > 3) loselife = true;
		}
		if (loselife) {
			p.lose_life();
			command << CMD_HITWALL << p.playerid << DELIMINATER;
			needsend = true;
			for (int i = 0; i < numplayers; i++) players[i]->draw(true);
		} else if (find(apples.begin(), apples.end(), p.currpos) != apples.end()) {
			Point newapplepos;
			do {
				newapplepos.x = rand() % 80;
				newapplepos.y = rand() % 50;
			} while (!FreePos(newapplepos));
			command << CMD_ATEAPPLE << p.playerid << ' ' << px << ' ' << py << ' ' << newapplepos.x << ' ' << newapplepos.y << DELIMINATER;
			replace(apples.begin(), apples.end(), p.currpos, newapplepos);
			if (!multiplayer) {
				p.score += 10;
				mainwindow->draw_apple(newapplepos);
			}
			needsend = true;
		}
		if (multiplayer && needsend) {
			send(servfd, command.str().c_str(), command.str().size(), 0);
		}
	}
	return true;
}

bool Server::Connect(const string &hostip)
{
	sockaddr_in servaddr;
	servfd = socket(AF_INET, SOCK_STREAM, 0);
	if ( servfd < 0 ) {
		return false;
	}
	memset(&servaddr, 0, sizeof(servaddr));
	servaddr.sin_family = AF_INET;
	servaddr.sin_port = htons(SPAKES_PORT);
	servaddr.sin_addr.s_addr = inet_addr(hostip.c_str());

	if (connect(servfd, (struct sockaddr *) &servaddr, sizeof(servaddr)) < 0) {
		closesocket(servfd);
		servfd = -1;
		return false;
	}

	return true;
}

bool Server::get_game_info()
{
	ostringstream command;
	command << CMD_GET_GAMEINFO << DELIMINATER;

	if (send(servfd, command.str().c_str(), command.str().size(), 0) < 0) return false;

	return true;

}

void Server::login_primary(const string &name, unsigned short port)
{
	ostringstream command;
	command << CMD_PRIMARY << name << ' ' << port << DELIMINATER;
	send(servfd, command.str().c_str(), command.str().size(), 0);
}

void Server::login_secondary(const string &name)
{
	ostringstream command;
	command << CMD_SECONDARY << ' ' << name << DELIMINATER;
	send(servfd, command.str().c_str(), command.str().size(), 0);
}

void Server::send_version_ok()
{
	ostringstream okstring;
	okstring << "Alright, you can let me in." << DELIMINATER;
	send(servfd, okstring.str().c_str(), okstring.str().size(), 0);
}

void Server::logout()
{
	if (servfd < 0) return;
	shutdown(servfd, 1);
	unsigned char ignore;
	while(read_byte(ignore) != 0);
	closesocket(servfd);
	servfd = -1;
}
