/*
    Spakes - a client/server Snakes-clone for up to 4 players
    Copyright (C) 2001, 2002, 2004  Markus Triska

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Contact me at triska@gmx.at
*/

#include <string>
#include <sstream>
#include <iostream>
#include <vector>
#include <fstream>

#include <stdlib.h>
#include <time.h>

#include "spakes.h"
#ifdef WINDOWS
#include <winsock.h>
#include "..\client\protocol.h"
#else
#include "../client/protocol.h"
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <stdio.h>
#include <netdb.h>
#include <linux/if.h>
#include <string.h>
#include <errno.h>
#endif

#include "init.h"
#include "socket.h"
#include "player.h"

using namespace std;

extern int numplayers;
extern int maxplayers;
extern Socket clientfd[4];
extern Player players[4];

extern void broadcast_cmd(int , ostringstream &, bool includesender = false);
extern void ByeByePlayer(int);

namespace init {

string localip;

void find_starting_position(int player, Point &p, int &dir)
{
	/* should use the symbolic DIR_... names from client/player.cpp here */
	switch(player) {
	case 0: p = Point(0, 25);  dir = 3; break;
	case 1: p = Point(79, 25); dir = 7; break;
	case 2: p = Point(40, 0);  dir = 5; break;
	case 3: p = Point(40, 49); dir = 1; break;
	}
}

int find_free_playernum()
{
	for (int i = 0; i < maxplayers; i++) {
		if (!players[i].ingame) return i;
	}
	return -1;
}


bool establish_connections()
{

#ifdef WINDOWS
  	struct hostent *hptr;
	char locname[100];
	char *hname;

	gethostname(locname, 99);
	if ( (hptr = gethostbyname(locname)) != NULL) {
		int i = 0;
		while(hptr->h_addr_list[i] != NULL) {
			in_addr addr;
			memcpy(&addr, hptr->h_addr_list[i], sizeof(struct in_addr));
			cout << "Local IP: " << inet_ntoa(addr) << endl;
			localip = inet_ntoa(addr);
			i++;
		}
	}
#endif

	int listenfd;

	if( (listenfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		cout << "Could not create socket. ";
#ifdef WINDOWS
		cout << "Error: " << WSAGetLastError() << endl;
#endif
		return false;
	}

	sockaddr_in listenaddr;
	memset(&listenaddr, 0, sizeof(listenaddr));
	listenaddr.sin_port = htons(SPAKES_PORT);
	listenaddr.sin_family = AF_INET;
	listenaddr.sin_addr.s_addr = htonl(INADDR_ANY);
	if (bind(listenfd, (sockaddr *) &listenaddr, sizeof(listenaddr)) < 0) {
		cout << "Could not bind port!" << endl;
		closesocket(listenfd);
		return false;
	}

#ifndef WINDOWS
	ifstream devices("/proc/net/dev");
	while (!devices.eof()) {
		string isdevice;
		devices >> isdevice;
		if (devices.eof()) break;
		if (isdevice[isdevice.size() - 1] == ':') {
			// this is a device (like "lo:" or "ppp0:")
			string ncolon(isdevice, 0, isdevice.size() - 1);
			struct ifreq i;
			strncpy(i.ifr_name, ncolon.c_str(), ncolon.size() + 1);
			if (ioctl(listenfd, SIOCGIFADDR, (int) &i) < 0) {
			//	cout << "ioctl() has failed for interface: " << ncolon << " with error:" << errno << endl;
			} else {
				sockaddr_in *sin = (sockaddr_in *) &i.ifr_addr;
				localip = inet_ntoa(sin->sin_addr);
				if (sin->sin_addr.s_addr != 0) {
					cout << "IP of interface " << ncolon << ": " << localip << endl;
				}
			}
		}
	}
	devices.close();
#endif

	if (listen(listenfd, 5) < 0) {
		cout << "LISTEN failed!" << endl;
		closesocket(listenfd);
		return false;
	}

	// Create apples
	srand(time(NULL));
	// Find 10 apple positions
	vector<Point> apples;
	for (int i = 0; i < 10; i++) {
		Point p;
		do {
			p.x = rand() % 80;
			p.y = rand() % 50;
		} while (find(apples.begin(), apples.end(), p) != apples.end());

		apples.push_back(p);
	}


	numplayers = 0;
	int currentplayer;
	while (numplayers < maxplayers) {
		sockaddr_in clientaddr;
#ifdef WINDOWS
		int clilen;
#else
		unsigned int clilen;
#endif
		clilen = sizeof(clientaddr);
		cout << "Waiting for player " << numplayers + 1 << "/" << maxplayers << "...";

		fd_set allset;
		int maxfd = listenfd;
		FD_ZERO(&allset);
		FD_SET(listenfd, &allset);
		for (int i = 0; i < maxplayers; i++) {
			if (players[i].primary && (players[i].ingame)) {
				FD_SET(clientfd[i].sockfd, &allset);
				if (clientfd[i].sockfd > maxfd) maxfd = clientfd[i].sockfd;
			}
		}

		if (select(maxfd + 1, &allset, NULL, NULL, NULL) < 0) {
			cout << "SELECT returned error." << endl;
			return false;
		}

		for (int i = 0; i < maxplayers; i++) {
			if (players[i].primary && FD_ISSET(clientfd[i].sockfd, &allset)) {
				string command;
				ostringstream reply;
				if (clientfd[i].Readline(command) == 0) {
					// Connection closed
					ByeByePlayer(i);
					numplayers--;
					if (players[i].secondarynum > -1) {
						numplayers--;
					}
					continue;
				}
				unsigned char com_comid;
				istringstream icommand(command);
				icommand >> com_comid;
				switch(com_comid) {
				case CMD_SECONDARY:
					// A primary player wants to register a secondary player
					currentplayer = find_free_playernum();
					cout << endl << "Player " << i + 1 << " is registering a secondary player." << endl;
					icommand >> players[currentplayer].name;
					Point p; int dir;
					find_starting_position(currentplayer, p, dir);

					reply << p.x << ' ' << p.y << ' ' << dir << DELIMINATER;

					players[currentplayer].beginpos = p;
					players[currentplayer].ingame = true;
					players[currentplayer].numlives = 3;
					players[currentplayer].primary = false;
					players[currentplayer].secondarynum = -1;

					// Inform all joint players
					// (but exclude the primary player,
					// player[i], who sent the request)

					ostringstream playerinfo;
					const char *thisplayername = players[currentplayer].name.c_str();
					playerinfo << SRV_REMSECONDARY << currentplayer << ' ' << thisplayername << ' ';
					playerinfo << p.x << ' ' << p.y << DELIMINATER;
					broadcast_cmd(i, playerinfo);

					// Send SRV_SECONDARY to the
					// primary player (player[i])
					ostringstream secplayerinfo;
					secplayerinfo << SRV_SECONDARY << currentplayer << ' ' << thisplayername << ' ' << p.x << ' ' << p.y << ' ' << dir << DELIMINATER;
					send(clientfd[i].sockfd, secplayerinfo.str().c_str(), secplayerinfo.str().size(), 0);
					players[i].secondarynum = currentplayer;					numplayers++;
					if (numplayers == maxplayers) goto enoughplayers;
					break;
				}
			}
		}


		if (FD_ISSET(listenfd, &allset)) {
			// A new primary player is trying to login
			currentplayer = find_free_playernum();
			if ( (clientfd[currentplayer].sockfd = accept(listenfd, (struct sockaddr *) &clientaddr, &clilen)) < 0) {
				cout << "ACCEPT returned error." << endl;
				return false;
			}
			cout << "accepted!" << endl;

			string command;
			clientfd[currentplayer].Readline(command);
			unsigned char com_comid;
			istringstream icommand(command);
			icommand >> com_comid;

			if (com_comid != CMD_PRIMARY) {
				cout << "Unexpected code" << endl;
			}
			string playername, clientreply;
			int jointplayer;
			ostringstream reply, serverver;

			sockaddr_in findclientip;
#ifdef WINDOWS
			int sizeofcip;
#else
			unsigned int sizeofcip;
#endif
			sizeofcip = sizeof(findclientip);
			if (getpeername(clientfd[currentplayer].sockfd, (struct sockaddr *) &findclientip, &sizeofcip) < 0) {
				cout << "Could not determine client IP!" << endl;
			}
			string clientip = inet_ntoa(findclientip.sin_addr);
			if (clientip == "127.0.0.1") clientip = localip;
			cout << "Client IP: " << clientip << endl;
			cout << "Checking client version..." << endl;
			// Send server version
			serverver << SRV_GAMEINFO << PROTOCOL_VERSION << DELIMINATER;
			send(clientfd[currentplayer].sockfd, serverver.str().c_str(), serverver.str().size(), 0);
			// wait for answer - if closing, reject
			if (clientfd[currentplayer].Readline(clientreply) == 0) {
				cout << "This login is rejected." << endl;
				closesocket(clientfd[currentplayer].sockfd);
				continue;
			} else {
			//	cout << "The client responded with: " << clientreply << endl;
			}

			icommand >> players[currentplayer].name;
			unsigned short port;
			icommand >> port;
			reply << SRV_LOGINREPLY << 1 << ' ' << currentplayer << ' ' << players[currentplayer].name.c_str() << ' ';

			// Find starting position and direction
			Point p; int dir;
			find_starting_position(currentplayer, p, dir);

			reply << p.x << ' ' << p.y << ' ' << dir << DELIMINATER;
			cout << "Client is logging in! " << endl;

			players[currentplayer].beginpos = p;
			players[currentplayer].ip = clientip;
			players[currentplayer].port = port;
			players[currentplayer].primary = true;
			players[currentplayer].numlives = 3;
			players[currentplayer].ingame = true;
			players[currentplayer].secondarynum = -1;

			send(clientfd[currentplayer].sockfd, reply.str().c_str(), reply.str().size(), 0);

			// Inform all other players joint already
			ostringstream playerinfo;
			const char *thisplayername = players[currentplayer].name.c_str();
			playerinfo << SRV_REMPRIMARY << currentplayer << ' ' << thisplayername << ' ';
			playerinfo << p.x << ' ' << p.y << ' ' << clientip.c_str() << ' ' << port << DELIMINATER;
			broadcast_cmd(currentplayer, playerinfo);
			// Inform this new player about the state of the others
			for ( jointplayer = 0; jointplayer < maxplayers; jointplayer++) {
				if (!players[jointplayer].ingame) continue;
				if (jointplayer == currentplayer) continue;

				ostringstream playerinfo1;
				const char *playername = players[jointplayer].name.c_str();
				if (players[jointplayer].primary) {
					playerinfo1 << SRV_REMPRIMARY;
				} else {
					playerinfo1 << SRV_REMSECONDARY;
				}
				playerinfo1 << jointplayer << ' ' << playername << ' ' ;
				playerinfo1 << players[jointplayer].beginpos.x << ' ' << players[jointplayer].beginpos.y;
				if (players[jointplayer].primary) {
					playerinfo1 << ' ' << players[jointplayer].ip.c_str() << ' ' << players[jointplayer].port;
				}
				playerinfo1 << DELIMINATER;
				send(clientfd[currentplayer].sockfd, playerinfo1.str().c_str(), playerinfo1.str().size(), 0);
			}

			// Send him the apples
			ostringstream appleinfo;
			for (vector<Point>::iterator i = apples.begin(); i != apples.end(); i++) {
				appleinfo << SRV_APPLE << i->x << ' ' << i ->y << DELIMINATER;
			}
			send(clientfd[currentplayer].sockfd, appleinfo.str().c_str(), appleinfo.str().size(), 0);

			numplayers++;
		}

	}
enoughplayers:
	closesocket(listenfd);
	return true;
}

bool initialize_game()
{
#ifdef WINDOWS
	Sleep(2000);
#else
	sleep(2);
#endif
	ostringstream gocmd;
	gocmd << SRV_GO << DELIMINATER;
	broadcast_cmd(0, gocmd, true);
	return true;
}

}; // End of namespace "init"
